#ifndef CFFGRAPH_H
#define CFFGRAPH_H

#include <fstream>
#include <vector>
#include <queue>


// Algorithms BFS Augmenting and Ford&Fulkerson's maximum flow are
// based on JAVA programcodes which are downloadable on the Internet



// define class Custom Ford-Fulkerson Graph
class CFFGraph
{

  private:
  
	  std::vector<bool> possibilityMatrix;

    unsigned int numberOfCities;
  	unsigned int numberOfWizards;
  	unsigned int numberOfRows;
  	
    struct coordinate
    {
      float x;
      float y;
    };
    
    bool isDistanceOK(coordinate a, coordinate b)
    {
      return ((a.x-b.x)*(a.x-b.x) + (a.y-b.y)*(a.y-b.y) <= 2500);
      // return (sqrt(pow((a.x-b.x), 2) + pow((a.y-b.y), 2)) <= 50);
    }
    
   	std::vector<int> BFSAugmenting();


  public:
  
		CFFGraph(const char* fileName);
		// we have a default destructor
  	int	FordFulkerson();
  	
}; // end define class


////////////////////////////////////////
CFFGraph::CFFGraph(const char* fileName)
{

  // FILEOPEN //
	std::ifstream input;
	input.open(fileName);
  if (input.fail())       // the file is not to open
  {
    std::cerr << std::endl << " Error: the file does not extist... " << std::endl;
    std::cout << " type any key and then <ENTER> to exit... ";
    char dummy;
    std::cin >> dummy;
    exit(2);
  }


  // FILEREAD //
	input >> numberOfCities;
	input >> numberOfWizards;
	numberOfRows = numberOfCities+numberOfWizards+2;

	std::vector<coordinate> wizardsAndCities(numberOfCities+numberOfWizards);
	for (unsigned int i=0; i<numberOfCities+numberOfWizards; i++)
	{
		input >> wizardsAndCities[i].x;
		input >> wizardsAndCities[i].y;
	}


  // create a matrix which contains the possibility of shows could be hold
	possibilityMatrix.resize(numberOfRows*numberOfRows);
	for (unsigned int i=0; i<numberOfCities; i++)
	{
		for (unsigned int j=numberOfCities; j<numberOfCities+numberOfWizards; j++)
		{
			if (isDistanceOK(wizardsAndCities[i], wizardsAndCities[j]))
			{
				possibilityMatrix[(i+1)*numberOfRows+j+1] = true;
      }
    }
	}

	for (unsigned int i=numberOfCities; i<numberOfRows-2; i++)
	{
		possibilityMatrix[(i+1)*numberOfRows] = true;
  }
	for (unsigned int i=1; i<numberOfCities+1; i++)
	{
		possibilityMatrix[(numberOfRows-1)*numberOfRows+i] = true;
  }

	input.close();
	
}; // end of constructor


//////////////////////////////////////////
std::vector<int> CFFGraph::BFSAugmenting()
{
	int sink   = 0;
	int source = numberOfRows-1;

	std::vector<int> c(numberOfRows);
	std::vector<int> d(numberOfRows, int(-1));
	std::vector<int> p(numberOfRows, int(-1));
	// I had to work with ints becouse of these

	c[source]=1;
  d[source]=0;

	std::queue<int> q;
	q.push(source);

	while (!q.empty() && source!=0)
	{
		source = q.front();
    q.pop();
		for (int i=0; i<numberOfRows; i++)
		{
			if (possibilityMatrix[source*numberOfRows+i])
			{
				if (c[i] == 0)
				{
					c[i] = 1;
					d[i] = d[source] + 1;
					p[i] = source;
					q.push(i);
				}
			}
		}
		c[source] = 2;
	}

	std::vector<int> path;

	if (d[sink] > 0)
	{
		source = sink;
		while (source!=numberOfRows-1)
		{
			path.push_back(source);
			source = p[source];
		}
		path.push_back(source);
	}

	return path;
}; // end of BFS Augmenting




/////////////////////////////
int CFFGraph::FordFulkerson()
{
	int u, v;
	
	std::vector<int> path;
	for (path=BFSAugmenting(); !path.empty(); path=BFSAugmenting())
	{
		for (int i=path.size()-1; i>0; i--)
		{
			u = path[i];
      v = path[i-1];
			possibilityMatrix[u*numberOfRows+v] = false;
			possibilityMatrix[v*numberOfRows+u] = true;
		}
	}
	
	int maxFlow = 0;
	for (int i=0; i<numberOfRows; i++)
	{
    if (possibilityMatrix[i])
    {
		  maxFlow++;
    }
  }

	return maxFlow;
}; // end of Ford&Fulkerson maxflow



#endif

